#include <stdio.h>
#include <stdlib.h>
#include <portaudio.h>

#include "ugens/sinosc.h"
#include "ugens/mult.h"
#include "ugens/add.h"
#include "ugens/expenv.h"
#include "ugens/linen.h"
#include "ugens/wshaper.h"

struct graph_node_s {
    void *ugen_instance;
    void (*tick)(void *);
    void (*destroy)(void *);
    struct graph_node_s *next;
};

typedef struct graph_node_s graph_node_t;

typedef struct
{
    graph_node_t *graphHead;
    double out1;
    double out2;
} paData;

static int paCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData )
{
    /* Cast data passed through stream to our structure. */
    paData *data = (paData*)userData;
    float *out = (float*)outputBuffer;
    float *in = (float*)inputBuffer;
    unsigned int i;
    (void) inputBuffer; /* Prevent unused variable warning. */
    for( i=0; i<framesPerBuffer; i++ ) {
        graph_node_t *node = data->graphHead;
        while (node) {
            node->tick(node->ugen_instance);
            node = node->next;
        }
        *out++ = data->out1;
        *out++ = data->out2;
    }
    return 0;
}



int main(void)
{
    PaError err;
    int sr = 44100;
    paData data;
    data.out1 = data.out2 = 0;

    double freq = 440;

//    graph_t *graph = create_graph();

//    graph_node_t *env = create_node(graph, "linen", sr, 1.0, 3, 1, 2);
//    graph_node_t *osc = create_node(graph, "sinosc", sr, freq, 0.5);
//    graph_node_t *shaper = create_node(graph, "wshaper");

    // Instantiate node
    graph_node_t *osc = (graph_node_t *) malloc(sizeof(graph_node_t));
    graph_node_t *shaper = (graph_node_t *) malloc(sizeof(graph_node_t));
    graph_node_t *env = (graph_node_t *) malloc(sizeof(graph_node_t));

    // Populate the node struct with instance and function pointer
    osc->ugen_instance = sinosc_construct(sr, freq, 0.5);
    osc->tick = sinosc_process;
    osc->destroy = sinosc_destroy;

    shaper->ugen_instance = wshaper_construct();
    shaper->tick = wshaper_process;
    shaper->destroy = wshaper_destroy;

    env->ugen_instance = linen_construct(sr, 1.0, 3, 1, 2);
    env->tick = linen_process;
    env->destroy = linen_destroy;

    // Build graph
    env->next = osc;
    osc->next = shaper;
    shaper->next = NULL;

    data.graphHead = env;

    // Connect outputs
    linen_set_output(env->ugen_instance, 0,
                     wshaper_get_input(shaper->ugen_instance, WSHAPER_MORPH_RATIO));
    sinosc_set_output(osc->ugen_instance, 0,
                      wshaper_get_input(shaper->ugen_instance, WSHAPER_IN));
    wshaper_set_output(shaper->ugen_instance, 0, &data.out1);

    // Set default values
    wshaper_set_param(shaper->ugen_instance, WSHAPER_DEPTH, 0.5);

    err = Pa_Initialize();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    PaStream *stream;
    err = Pa_OpenDefaultStream( &stream,
                                2, /* no input channels */
                                2, /* stereo output */
                                paFloat32, /* 32 bit floating point output */
                                sr,
                                32, /* frames per buffer */
                                paCallback, /* this is your callback function */
                                &data ); /*This is a pointer that will be passed to
   your callback*/
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_StartStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    Pa_Sleep(6*1000);

    err = Pa_StopStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err= Pa_CloseStream(stream);
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_Terminate();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    while (data.graphHead) {
        data.graphHead->destroy(data.graphHead->ugen_instance);
        graph_node_t *oldNode = data.graphHead;
        data.graphHead = data.graphHead->next;
        free(oldNode);
    }

    printf("Hello World!\n");
    return 0;
}

